@@ -33,6 +33,10 @@ typedef enum
   BREAK_PROHIBITED, /* no break, even if spaces intervene */
   BREAK_IF_SPACES,  /* "indirect break" (only if there are spaces) */
   BREAK_ALLOWED     /* "direct break" (can always break here) */
+  /* TR 14 has one more break-opportunity class,
+   * "indirect break opportunity for combining marks following a space"
+   * but we handle that inline in the code.
+   */
 } BreakOpportunity;
 
 enum
@@ -57,8 +61,10 @@ enum
   INDEX_BEFORE_AND_AFTER,
   INDEX_ZERO_WIDTH_SPACE,
   INDEX_COMBINING_MARK,
+  INDEX_WORD_JOINER,
 
   /* End of the table */
+
   INDEX_END_OF_TABLE,
 
   /* The following are not in the tables */
@@ -70,7 +76,8 @@ enum
   INDEX_SPACE,
   INDEX_COMPLEX_CONTEXT,
   INDEX_AMBIGUOUS,
-  INDEX_UNKNOWN
+  INDEX_UNKNOWN,
+  INDEX_NEXT_LINE
 };
 
 static BreakOpportunity row_OPEN_PUNCTUATION[INDEX_END_OF_TABLE] = {
@@ -78,7 +85,8 @@ static BreakOpportunity row_OPEN_PUNCTUATION[INDEX_END_OF_TABLE] = {
   BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
-  BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_CLOSE_PUNCTUATION[INDEX_END_OF_TABLE] = {
@@ -86,7 +94,8 @@ static BreakOpportunity row_CLOSE_PUNCTUATION[INDEX_END_OF_TABLE] = {
   BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_QUOTATION[INDEX_END_OF_TABLE] = {
@@ -94,7 +103,8 @@ static BreakOpportunity row_QUOTATION[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_NON_BREAKING_GLUE[INDEX_END_OF_TABLE] = {
@@ -102,7 +112,8 @@ static BreakOpportunity row_NON_BREAKING_GLUE[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_NON_STARTER[INDEX_END_OF_TABLE] = {
@@ -110,7 +121,8 @@ static BreakOpportunity row_NON_STARTER[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_EXCLAMATION[INDEX_END_OF_TABLE] = {
@@ -118,7 +130,8 @@ static BreakOpportunity row_EXCLAMATION[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_SYMBOL[INDEX_END_OF_TABLE] = {
@@ -126,15 +139,17 @@ static BreakOpportunity row_SYMBOL[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_INFIX_SEPARATOR[INDEX_END_OF_TABLE] = {
   BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_PREFIX[INDEX_END_OF_TABLE] = {
@@ -142,7 +157,8 @@ static BreakOpportunity row_PREFIX[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_IF_SPACES, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_POSTFIX[INDEX_END_OF_TABLE] = {
@@ -150,7 +166,8 @@ static BreakOpportunity row_POSTFIX[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_NUMERIC[INDEX_END_OF_TABLE] = {
@@ -158,7 +175,8 @@ static BreakOpportunity row_NUMERIC[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_ALPHABETIC[INDEX_END_OF_TABLE] = {
@@ -166,7 +184,8 @@ static BreakOpportunity row_ALPHABETIC[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_IDEOGRAPHIC[INDEX_END_OF_TABLE] = {
@@ -174,7 +193,8 @@ static BreakOpportunity row_IDEOGRAPHIC[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_INSEPARABLE[INDEX_END_OF_TABLE] = {
@@ -182,15 +202,17 @@ static BreakOpportunity row_INSEPARABLE[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_HYPHEN[INDEX_END_OF_TABLE] = {
   BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_AFTER[INDEX_END_OF_TABLE] = {
@@ -198,7 +220,8 @@ static BreakOpportunity row_AFTER[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_BEFORE[INDEX_END_OF_TABLE] = {
@@ -206,7 +229,8 @@ static BreakOpportunity row_BEFORE[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_BEFORE_AND_AFTER[INDEX_END_OF_TABLE] = {
@@ -214,7 +238,8 @@ static BreakOpportunity row_BEFORE_AND_AFTER[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity row_ZERO_WIDTH_SPACE[INDEX_END_OF_TABLE] = {
@@ -222,7 +247,8 @@ static BreakOpportunity row_ZERO_WIDTH_SPACE[INDEX_END_OF_TABLE] = {
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED, BREAK_ALLOWED,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_ALLOWED
 };
 
 static BreakOpportunity row_COMBINING_MARK[INDEX_END_OF_TABLE] = {
@@ -230,7 +256,17 @@ static BreakOpportunity row_COMBINING_MARK[INDEX_END_OF_TABLE] = {
   BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
   BREAK_ALLOWED, BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES,
   BREAK_ALLOWED, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
-  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_IF_SPACES
+  BREAK_ALLOWED, BREAK_ALLOWED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
+};
+
+static BreakOpportunity row_WORD_JOINER[INDEX_END_OF_TABLE] = {
+  BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_IF_SPACES, BREAK_IF_SPACES,
+  BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
+  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_IF_SPACES,
+  BREAK_IF_SPACES, BREAK_IF_SPACES, BREAK_PROHIBITED, BREAK_PROHIBITED,
+  BREAK_PROHIBITED
 };
 
 static BreakOpportunity *line_break_rows[INDEX_END_OF_TABLE] = {
@@ -253,7 +289,8 @@ static BreakOpportunity *line_break_rows[INDEX_END_OF_TABLE] = {
   row_BEFORE, /* INDEX_BEFORE */
   row_BEFORE_AND_AFTER, /* INDEX_BEFORE_AND_AFTER */
   row_ZERO_WIDTH_SPACE, /* INDEX_ZERO_WIDTH_SPACE */
-  row_COMBINING_MARK /* INDEX_COMBINING_MARK */
+  row_COMBINING_MARK, /* INDEX_COMBINING_MARK */
+  row_WORD_JOINER /* INDEX_WORD_JOINER */
 };
 
 /* Map GUnicodeBreakType to table indexes */
@@ -286,7 +323,9 @@ static int line_break_indexes[] = {
   INDEX_POSTFIX,
   INDEX_COMPLEX_CONTEXT,
   INDEX_AMBIGUOUS,
-  INDEX_UNKNOWN
+  INDEX_UNKNOWN,
+  INDEX_NEXT_LINE,
+  INDEX_WORD_JOINER
 };
 
 #define BREAK_INDEX(btype)                \
@@ -296,7 +335,7 @@ static int line_break_indexes[] = {
 #define BREAK_OP(before_type, after_type) \
          (BREAK_ROW (before_type)[BREAK_INDEX (after_type)])
 #define IN_BREAK_TABLE(btype)             \
-         (BREAK_INDEX(btype) < INDEX_END_OF_TABLE)
+         (btype <= G_N_ELEMENTS(line_break_indexes) && BREAK_INDEX(btype) < INDEX_END_OF_TABLE)
 
 /* Keep these in sync with the same macros in the test program */
 
@@ -328,6 +367,13 @@ static int line_break_indexes[] = {
 #define HIRAGANA(wc) ((wc) >= 0x3040 && (wc) <= 0x309F)
 #define KATAKANA(wc) ((wc) >= 0x30A0 && (wc) <= 0x30FF)
 
+#define LATIN(wc) (((wc) >= 0x0020 && (wc) <= 0x02AF) || ((wc) >= 0x1E00 && (wc) <= 0x1EFF))
+#define CYRILLIC(wc) (((wc) >= 0x0400 && (wc) <= 0x052F))
+#define GREEK(wc) (((wc) >= 0x0370 && (wc) <= 0x3FF) || ((wc) >= 0x1F00 && (wc) <= 0x1FFF))
+#define KANA(wc) ((wc) >= 0x3040 && (wc) <= 0x30FF)
+#define HANGUL(wc) ((wc) >= 0xAC00 && (wc) <= 0xD7A3)
+#define BACKSPACE_DELETES_CHARACTER(wc) (!LATIN (wc) && !CYRILLIC (wc) && !GREEK (wc) && !KANA(wc) && !HANGUL(wc))
+
 
 /* p. 132-133 of Unicode spec table 5-6 will help understand this */
 typedef enum
@@ -398,15 +444,16 @@ pango_default_break (const gchar   *text,
    */
 
   const gchar *next;
-  const gchar *end;
-  gint i = 0;
+  gint i;
   gunichar prev_wc;
   gunichar next_wc;
+  GUnicodeBreakType next_break_type;
   GUnicodeType prev_type;
   GUnicodeBreakType prev_break_type; /* skips spaces */
   gboolean prev_was_break_space;
   WordType current_word_type = WordNone;
   gunichar last_word_letter = 0;
+  gunichar base_character = 0;
   SentenceState sentence_state = STATE_SENTENCE_OUTSIDE;
   /* Tracks what will be the end of the sentence if a period is
    * determined to actually be a sentence-ending period.
@@ -414,61 +461,74 @@ pango_default_break (const gchar   *text,
   gint possible_sentence_end = -1;
   /* possible sentence break before Open* after a period-ended sentence */
   gint possible_sentence_boundary = -1;
+  gint n_chars;
 
   g_return_if_fail (text != NULL);
   g_return_if_fail (attrs != NULL);
 
-  if (length < 0)
-    length = strlen (text);
-  
-  next = text;
-  end = text + length;
-  
-  if (next == end)
-    return;
+  n_chars = g_utf8_strlen (text, length);
 
+  next = text;
+  
+  /* + 1 because of the extra newline we stick on the end */
+  if (attrs_len < n_chars + 1)
+    {
+      g_warning ("pango_default_break(): the array of PangoLogAttr passed in must have at least N+1 elements, if there are N characters in the text being broken");
+      return;
+    }
+      
   prev_type = (GUnicodeType) -1;
   prev_break_type = G_UNICODE_BREAK_UNKNOWN;
   prev_was_break_space = FALSE;
   prev_wc = 0;
 
-  next_wc = g_utf8_get_char (next);
+  if (n_chars)
+    {
+      next_wc = g_utf8_get_char (next);
+      g_assert (next_wc != 0);
+    }
+  else
+    next_wc = '\n';
 
-  g_assert (next_wc != 0);
+  next_break_type = g_unichar_break_type (next_wc);
 
-  while (next_wc != 0)
+  for (i = 0; i <= n_chars; i++)
     {
       GUnicodeType type;
       gunichar wc;
       GUnicodeBreakType break_type;
       BreakOpportunity break_op;
 
-      /* >, not >=, because of the extra newline we stick on the end */
-      if (i > attrs_len)
-        {
-          g_warning ("pango_default_break(): the array of PangoLogAttr passed in must have at least N+1 elements, if there are N characters in the text being broken");
-          return;
-        }
-      
       wc = next_wc;
+      break_type = next_break_type;
 
-      next = g_utf8_next_char (next);
-
-      if (next == end)
+      if (i == n_chars)
         {
-          /* This is how we fill in the last element (end position) of the
-           * attr array - assume there's a newline off the end of @text.
+          /*
+           * If we have already reached the end of @text g_utf8_next_char()
+           * may not increment next
            */
-          next_wc = '\n';
-        }
-      else if (next > end)
-        {
           next_wc = 0;
+	  next_break_type = G_UNICODE_BREAK_UNKNOWN;
         }
       else
         {
-          next_wc = g_utf8_get_char (next);
-          g_assert (next_wc != 0);
+          next = g_utf8_next_char (next);
+
+          if (i == n_chars - 1)
+            {
+              /* This is how we fill in the last element (end position) of the
+               * attr array - assume there's a newline off the end of @text.
+               */
+              next_wc = '\n';
+            }
+          else
+            {
+              next_wc = g_utf8_get_char (next);
+              g_assert (next_wc != 0);
+            }
+	  
+	  next_break_type = g_unichar_break_type (next_wc);
         }
 
       type = g_unichar_type (wc);
@@ -592,9 +652,15 @@ pango_default_break (const gchar   *text,
             }
         }
       
+      /* If this is a grapheme boundary, we have to decide if backspace
+       * deletes a character or the whole grapheme cluster */
+      if (attrs[i].is_cursor_position)
+        attrs[i].backspace_deletes_character = BACKSPACE_DELETES_CHARACTER (base_character);
+      else
+        attrs[i].backspace_deletes_character = FALSE;
+
       /* ---- Line breaking ---- */
 
-      break_type = g_unichar_break_type (wc);
       break_op = BREAK_ALREADY_HANDLED;
 
       g_assert (prev_break_type != G_UNICODE_BREAK_SPACE);
@@ -606,6 +672,13 @@ pango_default_break (const gchar   *text,
                                         * it's not a line break either
                                         */
         {
+	  /* space followed by a combining mark is handled
+	   * specially; (rule 7a from TR 14)
+	   */
+	  if (break_type == G_UNICODE_BREAK_SPACE &&
+	      next_break_type == G_UNICODE_BREAK_COMBINING_MARK)
+	    break_type = G_UNICODE_BREAK_IDEOGRAPHIC;
+		
           /* Unicode doesn't specify char wrap; we wrap around all chars
            * except where a line break is prohibited, which means we
            * effectively break everywhere except inside runs of spaces.
@@ -616,6 +689,7 @@ pango_default_break (const gchar   *text,
             {
             case G_UNICODE_BREAK_MANDATORY:
             case G_UNICODE_BREAK_LINE_FEED:
+            case G_UNICODE_BREAK_NEXT_LINE:
               attrs[i].is_line_break = TRUE;
               attrs[i].is_mandatory_break = TRUE;
               break;
@@ -638,9 +712,7 @@ pango_default_break (const gchar   *text,
               break;
 
             case G_UNICODE_BREAK_SURROGATE:
-              /* FIXME I have no clue what to do with these,
-               * but we should do something with them
-               */
+	      g_assert_not_reached ();
               break;
 
             case G_UNICODE_BREAK_AMBIGUOUS:
@@ -674,6 +746,7 @@ pango_default_break (const gchar   *text,
                 case G_UNICODE_BREAK_MANDATORY:
                 case G_UNICODE_BREAK_LINE_FEED:
                 case G_UNICODE_BREAK_CARRIAGE_RETURN:
+                case G_UNICODE_BREAK_NEXT_LINE:
                 case G_UNICODE_BREAK_SPACE:
                   /* These types all "pile up" at the end of lines and
                    * get elided.
@@ -699,9 +772,7 @@ pango_default_break (const gchar   *text,
                   break;
 
                 case G_UNICODE_BREAK_SURROGATE:
-                  /* FIXME this case needs to be handled
-                   */
-                  break_op = BREAK_IF_SPACES; /* not right at all */
+		  g_assert_not_reached ();
                   break;
 
                 default:
@@ -761,6 +832,7 @@ pango_default_break (const gchar   *text,
             case G_UNICODE_COMBINING_MARK:
             case G_UNICODE_ENCLOSING_MARK:
             case G_UNICODE_NON_SPACING_MARK:
+	    case G_UNICODE_FORMAT:
               /* nothing, we just eat these up as part of the word */
               break;
 
@@ -815,11 +887,9 @@ pango_default_break (const gchar   *text,
             default:
               /* Punctuation, control/format chars, etc. all end a word. */
               attrs[i].is_word_end = TRUE;
+	      current_word_type = WordNone;
               break;
             }
-
-          if (attrs[i].is_word_end)
-            current_word_type = WordNone;
         }
       else
         {
@@ -1271,7 +1341,13 @@ pango_default_break (const gchar   *text,
 
       prev_type = type;
       prev_wc = wc;
-      ++i;
+
+      /* wc might not be a valid unicode base character, but really all we
+       * need to know is the last non-combining character */
+      if (type != G_UNICODE_COMBINING_MARK && 
+          type != G_UNICODE_ENCLOSING_MARK && 
+          type != G_UNICODE_NON_SPACING_MARK)
+        base_character = wc;
     }
 }
 
@@ -1301,8 +1377,8 @@ pango_break (const gchar   *text,
     length = strlen (text);
 
   if (analysis->lang_engine &&
-      analysis->lang_engine->script_break)
-    (* analysis->lang_engine->script_break) (text, length, analysis, attrs, attrs_len);
+      PANGO_ENGINE_LANG_GET_CLASS (analysis->lang_engine)->script_break)
+    PANGO_ENGINE_LANG_GET_CLASS (analysis->lang_engine)->script_break (analysis->lang_engine, text, length, analysis, attrs, attrs_len);
   else
     pango_default_break (text, length, analysis, attrs, attrs_len);
 }
@@ -1523,4 +1599,3 @@ pango_get_log_attrs (const char    *text,
                  log_attrs + chars_broken,
                  attrs_len - chars_broken);
 }
-
