@@ -15,16 +15,16 @@
  *
  ******************************************************************/
 
-#include <freetype/tttags.h>
-
-#include <freetype/internal/ftstream.h>
-#include <freetype/internal/ftmemory.h>
-#include <freetype/internal/tttypes.h>
+#include "ftxopen.h"
+#include "ftxopenf.h"
 
 #include "fterrcompat.h"
 
-#include "ftxopen.h"
-#include "ftxopenf.h"
+#include FT_TRUETYPE_TAGS_H
+
+#include FT_INTERNAL_STREAM_H
+#include FT_INTERNAL_MEMORY_H
+#include FT_INTERNAL_TRUETYPE_TYPES_H
 
 #define TTAG_GDEF  FT_MAKE_TAG( 'G', 'D', 'E', 'F' )
 
@@ -132,6 +132,37 @@
   }
 #endif
 
+  EXPORT_FUNC
+  FT_Error  TT_New_GDEF_Table( FT_Face          face,
+			       TTO_GDEFHeader** retptr )
+  {
+    FT_Error         error;
+    FT_Memory        memory = face->memory;
+
+    TTO_GDEFHeader*  gdef;
+
+    if ( !retptr )
+      return TT_Err_Invalid_Argument;
+
+    if ( ALLOC( gdef, sizeof( *gdef ) ) )
+      return error;
+
+    gdef->memory = face->memory;
+
+    gdef->GlyphClassDef.loaded = FALSE;
+    gdef->AttachList.loaded = FALSE;
+    gdef->LigCaretList.loaded = FALSE;
+    gdef->MarkAttachClassDef_offset = 0;
+    gdef->MarkAttachClassDef.loaded = FALSE;
+
+    gdef->LastGlyph = 0;
+    gdef->NewGlyphClasses = NULL;
+
+    *retptr = gdef;
+
+    return TT_Err_Ok;
+  }
+
   EXPORT_FUNC
   FT_Error  TT_Load_GDEF_Table( FT_Face          face,
                                 TTO_GDEFHeader** retptr )
@@ -148,14 +179,12 @@
     if ( !retptr )
       return TT_Err_Invalid_Argument;
 
-    if (( error = tt_face->goto_table( tt_face, TTAG_GDEF, stream, 0 ) ))
+    if (( error = tt_face->goto_table( tt_face, TTAG_GDEF, stream, NULL ) ))
       return error;
 
-    if ( ALLOC( gdef, sizeof( *gdef ) ) )
+    if (( error = TT_New_GDEF_Table ( face, &gdef ) ))
       return error;
 
-    gdef->memory = face->memory;
-
     base_offset = FILE_Pos();
 
     /* skip version */
@@ -183,11 +212,9 @@
         goto Fail0;
       (void)FILE_Seek( cur_offset );
     }
-    else
-      gdef->GlyphClassDef.loaded = FALSE;
 
     if ( ACCESS_Frame( 2L ) )
-      goto Fail0;
+      goto Fail1;
 
     new_offset = GET_UShort();
 
@@ -204,11 +231,9 @@
         goto Fail1;
       (void)FILE_Seek( cur_offset );
     }
-    else
-      gdef->AttachList.loaded = FALSE;
 
     if ( ACCESS_Frame( 2L ) )
-      return error;
+      goto Fail2;
 
     new_offset = GET_UShort();
 
@@ -225,23 +250,30 @@
         goto Fail2;
       (void)FILE_Seek( cur_offset );
     }
-    else
-      gdef->LigCaretList.loaded = FALSE;
 
     /* OpenType 1.2 has introduced the `MarkAttachClassDef' field.  We
        first have to scan the LookupFlag values to find out whether we
-       must load it or not.  Here we only store the current file offset. */
+       must load it or not.  Here we only store the offset of the table. */
 
-    gdef->MarkAttachClassDef_offset = FILE_Pos();
-    gdef->MarkAttachClassDef.loaded = FALSE;
+    if ( ACCESS_Frame( 2L ) )
+      goto Fail3;
 
-    gdef->LastGlyph       = 0;
-    gdef->NewGlyphClasses = NULL;
+    new_offset = GET_UShort();
+
+    FORGET_Frame();
+
+    if ( new_offset )
+      gdef->MarkAttachClassDef_offset = new_offset + base_offset;
+    else
+      gdef->MarkAttachClassDef_offset = 0;
 
     *retptr = gdef;
 
     return TT_Err_Ok;
 
+  Fail3:
+    Free_LigCaretList( &gdef->LigCaretList, memory );
+    
   Fail2:
     Free_AttachList( &gdef->AttachList, memory );
 
@@ -266,6 +298,8 @@
     
     Free_NewGlyphClasses( gdef, memory );
 
+    FREE( gdef );
+
     return TT_Err_Ok;
   }
 
@@ -336,7 +370,7 @@
     FT_Memory memory = stream->memory;
     FT_Error  error;
 
-    FT_UShort         n, count;
+    FT_UShort         n, m, count;
     FT_ULong          cur_offset, new_offset, base_offset;
 
     TTO_AttachPoint*  ap;
@@ -392,8 +426,8 @@
     return TT_Err_Ok;
 
   Fail1:
-    for ( n = 0; n < count; n++ )
-      Free_AttachPoint( &ap[n], memory );
+    for ( m = 0; m < n; m++ )
+      Free_AttachPoint( &ap[m], memory );
 
     FREE( ap );
 
@@ -531,7 +565,7 @@
     FT_Memory memory = stream->memory;
     FT_Error  error;
 
-    FT_UShort        n, count;
+    FT_UShort        n, m, count;
     FT_ULong         cur_offset, new_offset, base_offset;
 
     TTO_CaretValue*  cv;
@@ -572,8 +606,8 @@
     return TT_Err_Ok;
 
   Fail:
-    for ( n = 0; n < count; n++ )
-      Free_CaretValue( &cv[n], memory );
+    for ( m = 0; m < n; m++ )
+      Free_CaretValue( &cv[m], memory );
 
     FREE( cv );
     return error;
@@ -609,7 +643,7 @@
     FT_Memory memory = stream->memory;
     FT_Error  error;
 
-    FT_UShort      n, count;
+    FT_UShort      m, n, count;
     FT_ULong       cur_offset, new_offset, base_offset;
 
     TTO_LigGlyph*  lg;
@@ -665,8 +699,8 @@
     return TT_Err_Ok;
 
   Fail1:
-    for ( n = 0; n < count; n++ )
-      Free_LigGlyph( &lg[n], memory );
+    for ( m = 0; m < n; m++ )
+      Free_LigGlyph( &lg[m], memory );
 
     FREE( lg );
 
@@ -712,7 +746,7 @@
 				   FT_UShort        glyphID,
 				   FT_UShort        index )
   {
-    FT_UShort              glyph_index, array_index;
+    FT_UShort              glyph_index, array_index, count;
     FT_UShort              byte, bits;
     
     TTO_ClassRangeRecord*  gcrr;
@@ -722,12 +756,13 @@
     if ( glyphID >= gdef->LastGlyph )
       return 0;
 
+    count = gdef->GlyphClassDef.cd.cd2.ClassRangeCount;
     gcrr = gdef->GlyphClassDef.cd.cd2.ClassRangeRecord;
     ngc  = gdef->NewGlyphClasses;
 
-    if ( glyphID < gcrr[index].Start )
+    if ( index < count && glyphID < gcrr[index].Start )
     {
-      array_index = 0;
+      array_index = index;
       if ( index == 0 )
         glyph_index = glyphID;
       else
@@ -739,7 +774,7 @@
       glyph_index = glyphID - gcrr[index].End - 1;
     }
 
-    byte = ngc[array_index][glyph_index / 4 + 1];
+    byte = ngc[array_index][glyph_index / 4];
     bits = byte >> ( 16 - ( glyph_index % 4 + 1 ) * 4 );
 
     return bits & 0x000F;
@@ -854,7 +889,7 @@
                                            FT_UShort*       class_array )
   {
     FT_UShort              start, curr_glyph, curr_class;
-    FT_UShort              n, count;
+    FT_UShort              n, m, count;
     FT_Error               error;
     FT_Memory              memory = gdef->memory;
 
@@ -965,7 +1000,7 @@
 
     if ( ALLOC_ARRAY( gdef->NewGlyphClasses,
                       gcd->cd.cd2.ClassRangeCount + 1, FT_UShort* ) )
-      goto Fail2;
+      goto Fail3;
 
     count = gcd->cd.cd2.ClassRangeCount;
     gcrr  = gcd->cd.cd2.ClassRangeRecord;
@@ -974,39 +1009,51 @@
     /* We allocate arrays for all glyphs not covered by the class range
        records.  Each element holds four class values.                  */
 
-    if ( gcrr[0].Start )
+    if ( count > 0 )
     {
-      if ( ALLOC_ARRAY( ngc[0], gcrr[0].Start / 4 + 1, FT_UShort ) )
-        goto Fail1;
-    }
+	if ( gcrr[0].Start )
+	{
+	  if ( ALLOC_ARRAY( ngc[0], ( gcrr[0].Start + 3 ) / 4, FT_UShort ) )
+	    goto Fail2;
+	}
 
-    for ( n = 1; n < count; n++ )
+	for ( n = 1; n < count; n++ )
+	{
+	  if ( gcrr[n].Start - gcrr[n - 1].End > 1 )
+	    if ( ALLOC_ARRAY( ngc[n],
+			      ( gcrr[n].Start - gcrr[n - 1].End + 2 ) / 4,
+			      FT_UShort ) )
+	      goto Fail1;
+	}
+
+	if ( gcrr[count - 1].End != num_glyphs - 1 )
+	{
+	  if ( ALLOC_ARRAY( ngc[count],
+			    ( num_glyphs - gcrr[count - 1].End + 2 ) / 4,
+			    FT_UShort ) )
+	      goto Fail1;
+	}
+    }
+    else if ( num_glyphs > 0 )
     {
-      if ( gcrr[n].Start - gcrr[n - 1].End > 1 )
-        if ( ALLOC_ARRAY( ngc[n],
-                          ( gcrr[n].Start - gcrr[n - 1].End - 1 ) / 4 + 1,
-                          FT_UShort ) )
-          goto Fail1;
+	if ( ALLOC_ARRAY( ngc[count],
+			  ( num_glyphs + 3 ) / 4,
+			  FT_UShort ) )
+	    goto Fail2;
     }
-
-    if ( gcrr[count - 1].End != num_glyphs - 1 )
-    {
-      if ( ALLOC_ARRAY( ngc[count],
-                        ( num_glyphs - gcrr[count - 1].End - 1 ) / 4 + 1,
-                        FT_UShort ) )
-        goto Fail1;
-    }
-
+	
     gdef->LastGlyph = num_glyphs - 1;
 
     gdef->MarkAttachClassDef_offset = 0L;
     gdef->MarkAttachClassDef.loaded = FALSE;
 
+    gcd->loaded = TRUE;
+
     return TT_Err_Ok;
 
   Fail1:
-    for ( n = 0; n < count; n++ )
-      FREE( ngc[n] );
+    for ( m = 0; m < n; m++ )
+      FREE( ngc[m] );
 
   Fail2:
     FREE( gdef->NewGlyphClasses );
@@ -1047,7 +1094,7 @@
     FT_Error               error;
     FT_UShort              class, new_class, index;
     FT_UShort              byte, bits, mask;
-    FT_UShort              array_index, glyph_index;
+    FT_UShort              array_index, glyph_index, count;
 
     TTO_ClassRangeRecord*  gcrr;
     FT_UShort**            ngc;
@@ -1088,12 +1135,13 @@
       return TT_Err_Invalid_Argument;
     }
 
+    count = gdef->GlyphClassDef.cd.cd2.ClassRangeCount;
     gcrr = gdef->GlyphClassDef.cd.cd2.ClassRangeRecord;
     ngc  = gdef->NewGlyphClasses;
 
-    if ( glyphID < gcrr[index].Start )
+    if ( index < count && glyphID < gcrr[index].Start )
     {
-      array_index = 0;
+      array_index = index;
       if ( index == 0 )
         glyph_index = glyphID;
       else
@@ -1105,7 +1153,7 @@
       glyph_index = glyphID - gcrr[index].End - 1;
     }
 
-    byte  = ngc[array_index][glyph_index / 4 + 1];
+    byte  = ngc[array_index][glyph_index / 4];
     bits  = byte >> ( 16 - ( glyph_index % 4 + 1 ) * 4 );
     class = bits & 0x000F;
 
@@ -1116,8 +1164,8 @@
       bits = new_class << ( 16 - ( glyph_index % 4 + 1 ) * 4 );
       mask = ~( 0x000F << ( 16 - ( glyph_index % 4 + 1 ) * 4 ) );
 
-      ngc[array_index][glyph_index / 4 + 1] &= mask;
-      ngc[array_index][glyph_index / 4 + 1] |= bits;
+      ngc[array_index][glyph_index / 4] &= mask;
+      ngc[array_index][glyph_index / 4] |= bits;
     }
 
     return TT_Err_Ok;
@@ -1125,27 +1173,53 @@
 
 
   FT_Error  Check_Property( TTO_GDEFHeader*  gdef,
-                            FT_UShort        index,
+			    OTL_GlyphItem    gitem,
                             FT_UShort        flags,
                             FT_UShort*       property )
   {
     FT_Error  error;
 
-
     if ( gdef )
     {
-      error = TT_GDEF_Get_Glyph_Property( gdef, index, property );
-      if ( error )
-        return error;
+      FT_UShort basic_glyph_class;
+      FT_UShort desired_attachment_class;
 
-      /* This is OpenType 1.2 */
+      if ( gitem->gproperties == OTL_GLYPH_PROPERTIES_UNKNOWN )
+      {
+	error = TT_GDEF_Get_Glyph_Property( gdef, gitem->gindex, &gitem->gproperties );
+	if ( error )
+	  return error;
+      }
 
-      if ( flags & IGNORE_SPECIAL_MARKS )
-        if ( (flags & 0xFF00) != *property )
-          return TTO_Err_Not_Covered;
+      *property = gitem->gproperties;
 
-      if ( flags & *property )
-        return TTO_Err_Not_Covered;
+      /* If the glyph was found in the MarkAttachmentClass table,
+       * then that class value is the high byte of the result,
+       * otherwise the low byte contains the basic type of the glyph
+       * as defined by the GlyphClassDef table.
+       */
+      if ( *property & IGNORE_SPECIAL_MARKS  )
+	basic_glyph_class = TTO_MARK;
+      else
+	basic_glyph_class = *property;
+
+      /* Return Not_Covered, if, for example, basic_glyph_class
+       * is TTO_LIGATURE and LookFlags includes IGNORE_LIGATURES
+       */
+      if ( flags & basic_glyph_class )
+	return TTO_Err_Not_Covered;
+      
+      /* The high byte of LookupFlags has the meaning
+       * "ignore marks of attachment type different than
+       * the attachment type specified."
+       */
+      desired_attachment_class = flags & IGNORE_SPECIAL_MARKS;
+      if ( desired_attachment_class )
+      {
+	if ( basic_glyph_class == TTO_MARK &&
+	     *property != desired_attachment_class )
+	  return TTO_Err_Not_Covered;
+      }
     }
 
     return TT_Err_Ok;
